__copyright__ = 'Copyright 2020-2021, The RADICAL-Cybertools Team'
__license__   = 'MIT'


NAME_MESSAGE = "Valid object names can contain letters, numbers and '.'. " \
               "Any other character will not be allowed in versions >= 1.5.10"
